"""
PodIO
https://github.com/AIDASoft/podio.git
"""

import os

from ejpm.engine.env_gen import Set, Append, Prepend
from ejpm.engine.git_cmake_recipe import GitCmakeRecipe


class PodioRecipe(GitCmakeRecipe):
    """Provides data for building and installing Genfit framework
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self):
        """
        Installs Genfit track fitting framework
        """

        # Set initial values for parent class and self
        super(PodioRecipe, self).__init__('podio')                        # This name will be used in ejpm commands
        self.config['branch'] = 'v00-13'                                # The branch or tag to be cloned (-b flag)
        self.config['repo_address'] = 'https://github.com/AIDASoft/podio.git'    # Repo address
        self.config['cmake_flags'] = '-DBUILD_TESTING=ON'
        self.config['cxx_standard'] = 17

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        path = data['install_path']

        yield Set('ACTS_DIR', path)

        # it could be lib or lib64. There are bugs on different platforms (RHEL&centos and WSL included)
        # https://stackoverflow.com/questions/46847939/config-site-for-vendor-libs-on-centos-x86-64
        # https: // bugzilla.redhat.com / show_bug.cgi?id = 1510073

        import platform
        if platform.system() == 'Darwin':
            if os.path.isdir(os.path.join(path, 'lib64')):
                yield Append('DYLD_LIBRARY_PATH', os.path.join(path, 'lib64'))
            else:
                yield Append('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

        if os.path.isdir(os.path.join(path, 'lib64')):
            yield Append('LD_LIBRARY_PATH', os.path.join(path, 'lib64'))
        else:
            yield Append('LD_LIBRARY_PATH', os.path.join(path, 'lib'))

        yield Append('CMAKE_PREFIX_PATH', os.path.join(path, 'lib', 'cmake', 'podio'))


    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "python-jinja2 python-yaml",
            'centos': "",
            'centos8': "",
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }


import importlib
import pkgutil
import ejpm.recipes
from ejpm.engine.git_cmake_recipe import GitCmakeRecipe

from ejpm.engine.recipe import Recipe


def import_all_submodules(modules_dir, package_name):
    # >oO debug   for (module_loader, name, ispkg) in pkgutil.iter_modules([modules_dir]):
    # >oO debug       print("module_loader {}, name {}, ispkg {}".format(module_loader, name, ispkg))

    # parent_module =  imp.find_module('packets', path)
    for (module_loader, name, ispkg) in pkgutil.iter_modules([modules_dir]):
        importlib.import_module('.' + name, package_name)
        # module = importlib.import_module('.' + name, package_name)
        # >oO debug   print(module)


class InstallationRequest(object):
    """ This class is assumed to reflect a user requests for a packet installation"""

    def __init__(self, recipe, mode, config, just_explain=False, deps_only=4):
        assert isinstance(recipe, Recipe)

        self.name = recipe.name   # Packet name
        self.recipe = recipe
        self.mode = mode
        self.just_explain = just_explain
        self.config_overrides = config
        self.deps_only = deps_only

    def update_installer_config(self):
        self.recipe.config.update(self.config_overrides)

    def __repr__(self):
        return "InstallationRequest '{}'".format(self.name)


class RecipeManager(object):

    class Config(object):
        dir = ""

    def __init__(self):
        # But now we just import and create them manually
        self.recipes_by_name = {}
        self.env_generators = {}

        # The next are collection of requirements for different operating systems
        # The type is map to have requirements by packets, i.e.:
        #     self.centos_required_packets['ejana'] - list for ejana
        self.os_deps_by_name = {}

    def load_installers(self, modules_dir="", package_name=""):

        if not modules_dir:
            modules_dir = ejpm.recipes.__path__[0]
        if not package_name:
            package_name = ejpm.recipes.__name__

        # We need to import submodules at least once to get __submodules__() function work later
        import_all_submodules(modules_dir, package_name)

        # Create all subclasses of Recipe and GitCmakeRecipe and add here
        classes = [cls for cls in Recipe.__subclasses__() + GitCmakeRecipe.__subclasses__() if cls != GitCmakeRecipe]

        for cls in classes:
            if cls == GitCmakeRecipe:
                continue
            installer = cls()

            # Add installer 'by name'
            self.recipes_by_name[installer.name] = installer

            # Add environment generator to env_generators map
            if hasattr(installer, 'gen_env'):
                self.env_generators[installer.name] = installer.gen_env

            # Add list of dependencies
            self.add_os_deps(installer)

    def add_os_deps(self, installer):
        """Adds os dependencies to global os_dependencies_by_name map"""

        # First, we add default structure with empty deps
        result = {'required': {'ubuntu': '', 'centos': ''},
                  'optional': {'ubuntu': '', 'centos': ''}}

        # Then we check if installer defines its dependencies
        if hasattr(installer, 'os_dependencies'):
            result.update(installer.os_dependencies)

        # Set the result by installer name
        self.os_deps_by_name[installer.name] = result

    def get_installation_chain_names(self, main_recepie_name, deps_only=False):
        """
        Returns name of the package + dependencies ejpm can install
        so it is like: ['CLHEP', 'root', ..., 'ejana'] for installer_name=ejana
        it is single: ['CLHEP'] for installer_name='CLHEP'

        :param deps_only: get only names of dependencies not packet name included
        :param main_recepie_name: name of packet like 'ejana'
        :return: list with dependencies names and installer name itself
        """

        deps = self.recipes_by_name[main_recepie_name].required_deps

        if deps_only:
            return deps

        # If we install just a single packet desired_names a single name
        return deps + [main_recepie_name] if deps else [main_recepie_name]

    def gen_shell_env_text(self, name_data, shell='bash'):
        """Generates a text that sets environment for a given shell """

        output = ""     # a string holding the result

        # Go through provided name-path pairs:
        for name, data in name_data.items():

            # if some packet has no data, or there is no environ generator for it, we skip it
            if not data or name not in self.env_generators.keys():
                continue

            env_gen = self.env_generators[name]

            output += "\n"
            output += "# =============================\n"
            output += "# {}\n".format(name)
            output += "# =============================\n"

            # env_gen(data) provides environment manipulation instructions based on the given data
            steps = env_gen(data)
            if not steps:
                continue
            for step in env_gen(data):
                output += step.gen_csh() if shell == 'csh' else step.gen_bash()   # bash or csh?
                output += '\n'
        return output

    def gen_bash_env_text(self, name_data):
        """Generates a text that sets environment for bash shell """
        return self.gen_shell_env_text(name_data, shell='bash')

    def gen_csh_env_text(self, name_data):
        """Generates a text that sets environment for csh/tcsh shell """
        return self.gen_shell_env_text(name_data, shell='csh')

    def update_python_env(self, name_data):
        """Updates python environment according to (name,paths) pairs"""

        # Go through provided name-path pairs:
        for name, data in name_data.items():

            # If we have a generator for this program and installation data
            if data and name in self.env_generators.keys():
                env_gens = self.env_generators[name]
                for env_gen in env_gens(data):          # Go through 'environment generators' look engine/env_gen.py
                    env_gen.update_python_env()         # Do environment update



"""
This is a generic recepie for git + cmake project

"""

import os

from ejpm.engine.commands import run, workdir
from ejpm.engine.env_gen import Set, Append
from ejpm.engine.recipe import Recipe


class GitCmakeRecipe(Recipe):
    """Provides data for building and installing Genfit framework
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self, name):
        """Generic init"""

        super(GitCmakeRecipe, self).__init__(name)

        # Set initial values for parent class and self
        self.config['branch'] = 'master'
        self.config['repo_address'] = ''
        self.config['cmake_flags'] = ''
        self.config['cmake_custom_flags'] = ''

    def setup(self, db):
        """Sets all variables like source dirs, build dirs, etc"""

        #
        # use_common_dirs_scheme sets standard package variables:
        # source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
        # build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
        # install_path = {app_path}/root-{version}         # Where the binary installation is
        self.use_common_dirs_scheme()

        #
        # Check we have a repo address
        if not self.config['repo_address']:
            raise ValueError("config 'repo_address' must be set. "
                             "Either in recepie or later with 'ejpm config <packet> repo_address=<value>'")

        #
        # Git download link. Clone with shallow copy
        self.clone_command = "git clone --depth 1 -b {branch} {repo_address} {source_path}"\
            .format(**self.config)

        # cmake command:
        # the  -Wno-dev  flag is to ignore the project developers cmake warnings for policy CMP0075
        self.build_cmd = "cmake -Wno-dev " \
                         "-DCMAKE_INSTALL_PREFIX={install_path} -DCMAKE_CXX_STANDARD={cxx_standard} " \
                         "{cmake_flags} {cmake_custom_flags} {source_path}" \
                         "&& cmake --build . -- -j {build_threads}" \
                         "&& cmake --build . --target install" \
                         .format(**self.config)

    def step_install(self):
        self.step_clone()
        self.step_build()

    def step_clone(self):
        """Clones JANA from github mirror"""

        # Check the directory exists and not empty
        if os.path.exists(self.source_path) and os.path.isdir(self.source_path) and os.listdir(self.source_path):
            # The directory exists and is not empty. Nothing to do
            return
        else:
            # Create the directory
            run('mkdir -p {}'.format(self.source_path))

        # Execute git clone command
        run(self.clone_command)

    def step_build(self):
        """Builds JANA from the ground"""

        # Create build directory
        run('mkdir -p {}'.format(self.build_path))

        # go to our build directory
        workdir(self.build_path)

        # run scons && scons install
        run(self.build_cmd)

    def step_reinstall(self):
        """Delete everything and start over"""

        # clear sources directories if needed
        run('rm -rf {}'.format(self.app_path))

        # Now run build root
        self.step_install()


import click

from ejpm.engine.api import pass_ejpm_context, EjpmApi
from ejpm.engine.output import markup_print as mprint


# @click.group(invoke_without_command=True)
@click.command()
@click.argument('packet_name', nargs=1)
@click.argument('install_path', nargs=1)
@pass_ejpm_context
@click.pass_context
def set(ctx, ectx, packet_name, install_path):
    """Sets packets"""

    assert isinstance(ectx, EjpmApi)

    # We need DB ready for this cli command
    ectx.ensure_db_exists()

    # Check that the packet name is from known packets
    ectx.ensure_installer_known(packet_name)

    # update_install will add or update the packet install. We set it active as it make sense...
    from ejpm.engine.db import IS_ACTIVE
    ectx.db.update_install(packet_name, install_path, {IS_ACTIVE: True})
    ectx.db.save()

    # Update environment scripts
    mprint("Updating environment script files...\n")
    ectx.save_default_bash_environ()
    ectx.save_default_csh_environ()



"""
This file provides information of how to build and configure Eic-smear framework:
https://gitlab.com/eic/eic-smear

"""

import os

from ejpm.engine.commands import run, workdir
from ejpm.engine.env_gen import Set, Append, Prepend
from ejpm.engine.git_cmake_recipe import GitCmakeRecipe


class EicSmearRecipe(GitCmakeRecipe):
    """Provides data for building and installing Genfit framework
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self):
        """"""
        # Set initial values for parent class and self
        super(EicSmearRecipe, self).__init__('eic-smear')
        self.clone_command = ''             # will be set by self.set_app_path
        self.build_cmd = ''                 # will be set by self.set_app_path
        self.required_deps = ['root']
        self.config['branch'] = '1.0.4-fix1'
        self.config['repo_address'] = 'https://gitlab.com/eic/eic-smear.git'

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""

        install_path = data['install_path']
        yield Set('EIC_SMEAR_HOME', install_path)
        yield Append('ROOT_INCLUDE_PATH', os.path.join(install_path, '/include'))

        lib_path = os.path.join(install_path, 'lib')  # on some platforms
        lib64_path = os.path.join(install_path, 'lib64')  # on some platforms

        if os.path.isdir(lib64_path):
            yield Prepend('LD_LIBRARY_PATH', lib64_path)
        else:
            yield Prepend('LD_LIBRARY_PATH', lib_path)

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "",
            'centos': ""
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }



"""
This file provides information of how to build and configure VMC framework:
Geometry conversion tool, providing conversion between Geant4 and ROOT TGeo geometry models.

https://github.com/vmc-project/vgm


"""

import os

from ejpm.engine.commands import run, workdir
from ejpm.engine.env_gen import Set, Append
from ejpm.engine.git_cmake_recipe import GitCmakeRecipe
from ejpm.engine.recipe import Recipe


class VgmRecipe(GitCmakeRecipe):
    """Provides data for building and installing Geant4 framework
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self):
        """
        Installs Genfit track fitting framework
        """

        # Set initial values for parent class and self
        super(VgmRecipe, self).__init__('vgm')
        self.config['branch'] = 'v4-5'
        self.config['cmake_flags'] = '-DWITH_EXAMPLES=0'
        self.config['repo_address'] = 'https://github.com/vmc-project/vgm'

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""

        install_path = data['install_path']
        yield Set('VGM_DIR', install_path)

        import platform
        if platform.system() == 'Darwin':
            yield Append('DYLD_LIBRARY_PATH', os.path.join(install_path, 'lib'))
            yield Append('DYLD_LIBRARY_PATH', os.path.join(install_path, 'lib64'))

        yield Append('LD_LIBRARY_PATH', os.path.join(install_path, 'lib'))
        yield Append('LD_LIBRARY_PATH', os.path.join(install_path, 'lib64'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "",
            'centos': ""
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }



class ConfigNamespace:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        keys = sorted(self.__dict__)
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def update(self, other):
        if isinstance(other, ConfigNamespace):
            self.__dict__.update(other.__dict__)
        elif isinstance(other, dict):
            self.__dict__.update(other)


import os
import subprocess
import sys
import click

executed_commands = []


class Command(object):
    """Base abstract class responsible for any command like run, env, etc"""

    def __init__(self):
        self.is_executed = False

    def execute(self):
        """executes the command. Must be implemented in inherited classes"""
        raise NotImplementedError()

    def __str__(self):
        return "Abstract command"


class RunCommand(Command):
    """Runs an OS command"""

    def __init__(self, args):
        super(RunCommand, self).__init__()
        self.args = args
        self.return_code = 1000000

    def execute(self):
        """executes system command"""

        # Print executing command
        click.secho("EXECUTING:", fg='blue', bold=True)
        click.echo(self.args)

        # Execute the command
        self.return_code = subprocess.call(self.args, stdout=sys.stdout, stderr=sys.stderr, shell=True)
        self.is_executed = True

        # Print the result
        click.secho("Execution done. ", fg='blue', bold=True, nl=False)
        click.echo("Return code = ", nl=False)
        click.secho("%i" % self.return_code, fg='red' if self.return_code else 'green', bold=True)
        click.echo()


class WorkDirCommand(Command):
    """Changes current working directory"""

    def __init__(self, path):
        super(WorkDirCommand, self).__init__()
        self.path = path

    def execute(self):
        """change directory according to path"""
        click.secho("WORKDIR: ", fg='blue', bold=True, nl=False)
        click.echo(self.path)
        click.echo()

        os.chdir(self.path)
        self.is_executed = True


class EnvironmentCommand(Command):
    """Sets environment variable"""

    def __init__(self, name, value):
        super(EnvironmentCommand, self).__init__()
        self.name = name
        self.value = value

    def execute(self):
        """ Set environment variable"""

        click.secho("ENV: ", fg='blue', bold=True, nl=False)
        click.echo("%s = %s" % (self.name, self.value))
        click.echo()
        os.environ[self.name] = self.value
        self.is_executed = True


def run(args):
    """Runs an OS command"""

    command = RunCommand(args)
    _execute_command(command)

    # check run command is ended with return code 0
    if command.return_code != 0:
        click.secho("ERROR", fg='red', bold=True, nl=True)
        click.echo(": Command return code != 0. COMMAND CONTENT:")
        click.echo(command.args)
        click.echo()

        raise OSError()


def workdir(path):
    """Changes working directory """

    _execute_command(WorkDirCommand(path))


def env(name, value):
    """Sets environment variable"""

    _execute_command(EnvironmentCommand(name, value))


def _execute_command(command):
    """Executes any command class"""

    assert isinstance(command, Command)  # sanity check
    command.execute()                    # execute the command!

    # add executed command to the list
    global executed_commands
    executed_commands.append(command)


def is_not_empty_dir(path):
    """Checks the path is existing directory and it is not empty"""

    return os.path.exists(path) and os.path.isdir(path) and os.listdir(path)




"""
This file provides information of how to build and configure CLHEP library:
https://gitlab.cern.ch/CLHEP/CLHEP

"""

import os

from ejpm.engine.env_gen import Set, Append, Prepend
from ejpm.engine.git_cmake_recipe import GitCmakeRecipe


class ClhepRecipe(GitCmakeRecipe):
    """Provides data for building and installing CLHEP framework
    """

    def __init__(self):
        # Set initial values for parent class and self
        super(ClhepRecipe, self).__init__('clhep')
        self.clone_command = ''             # will be set by self.set_app_path
        self.build_cmd = ''                 # will be set by self.set_app_path
        self.config['branch'] = 'master'
        self.config['repo_address'] = 'https://gitlab.cern.ch/CLHEP/CLHEP.git'

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""

        path = data['install_path']
        lib_path = os.path.join(path, 'lib')
        include_path = os.path.join(path, 'include')
        bin_path = os.path.join(path, 'bin')

        yield Set('CLHEP', path)
        yield Set('CLHEP_BASE_DIR', path)                  # Some system look for CLHEP this way
        yield Set('CLHEP_INCLUDE_DIR', include_path)  # or /usr/include/CLHEP/
        yield Set('CLHEP_LIB_DIR', lib_path)

        yield Prepend('PATH', bin_path)  # to make available clhep-config and others
        yield Prepend('LD_LIBRARY_PATH', lib_path)

        # set DYLD_LIBRARY_PATH for mac
        import platform
        if platform.system() == 'Darwin':
            yield Append('DYLD_LIBRARY_PATH', lib_path)

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "cmake",
            'centos': "cmake"
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }



"""
This file provides information of how to build and configure JANA2 packet:
https://github.com/JeffersonLab/JANA2

"""

import os

from ejpm.engine.commands import run, workdir, env
from ejpm.engine.env_gen import Prepend, Set, Append
from ejpm.engine.git_cmake_recipe import GitCmakeRecipe
from ejpm.engine.recipe import Recipe


class JanaRecipe(GitCmakeRecipe):
    """Provides data for building and installing JANA2 framework

    PacketInstallationInstruction is located in recipe.py and contains the next standard package variables:


    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self):
        super(JanaRecipe, self).__init__('jana')
        self.config['branch'] = 'v2.0.2'
        self.config['repo_address'] = 'https://github.com/JeffersonLab/JANA2.git'

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""

        install_path = data['install_path']

        yield Set('JANA_HOME', install_path)
        yield Append('JANA_PLUGIN_PATH', '$JANA_HOME/plugins')
        yield Prepend('PATH', '$JANA_HOME/bin')

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "scons libxerces-c-dev curl python3-dev",
            'centos': "scons xerces-c-devel curl",
            'centos8': "python3-scons xerces-c-devel curl"
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }



import os
import click
import copy

from ejpm.engine.api import pass_ejpm_context, EjpmApi
from ejpm.engine.db import PacketStateDatabase, BUILT_WITH_CONFIG
from ejpm.engine.output import markup_print as mprint
from ejpm.engine.recipe_manager import RecipeManager, InstallationRequest


@click.command()
@click.option('--missing', 'dep_mode', flag_value='missing', help="Installs only missing dependencies", default=True)
@click.option('--single', 'dep_mode', flag_value='single', help="Installs only this package")
@click.option('--force', 'dep_mode', flag_value='single', help="Force installation of a single package (same as --single)")
@click.option('--all', 'dep_mode', flag_value='all', help="Installs all dependencies by ejpm")
@click.option('--path', 'install_path', default='', help="Is not implemented")
@click.option('--build-threads', '-j', 'build_threads', default=0, help="Build threads count")
@click.option('--explain', 'just_explain', default=False, is_flag=True, help="Prints what is to be installed (but do nothing)")
@click.option('--deps-only', 'deps_only', default=False, is_flag=True, help="Installs only dependencies but not the packet itself")
@click.argument('names', nargs=-1)
@pass_ejpm_context
@click.pass_context
def install(ctx, ectx, dep_mode, names, install_path="", build_threads=4, just_explain=False, deps_only=False):
    """Installs packets (and all dependencies)

    \b
    Examples:
      > ejpm install ejana --missing    # install ejana and all missing dependencies
      > ejpm install rave --single      # install just rave package, dependencies are not checked
      > ejpm install ejana --all        # install all ejana dependencies by ejpm
                                        # even if user pointed some deps to external places

    \b
    --explain flag may be used to see what dependencies packet has and what is missing

      > ejpm install ejana --missing --explain   # print what to be installed but not install

    """

    db = ectx.db
    pm = ectx.pm
    assert isinstance(ectx, EjpmApi)
    assert isinstance(db, PacketStateDatabase)
    assert isinstance(pm, RecipeManager)

    # Check if packet_name is all, missing or for known packet
    for name in names:
        ectx.ensure_installer_known(name)

    # Ok, looks like we are going to install something

    # If no db...
    if not db.exists():
        mprint("<green>creating database...</green>")
        db.save()

    # Lets check if we have top_dir
    if not db.top_dir:
        _print_help_no_top_path()
        raise click.Abort()

    # Install packets
    # set the tag we want to install
    requests = []
    for name in names:
        # make config overrides
        config = {}
        config.update(db.get_global_config())
        config.update(db.get_config(name))
        if build_threads:
            config['build_threads'] = build_threads

        # make installation request and add to the list
        request = InstallationRequest(pm.recipes_by_name[name], dep_mode, config, just_explain, deps_only)
        requests.append(request)

    for request in requests:
        _install_with_deps(ectx, request)

    # Update environment scripts if it is not just an explanation
    if not just_explain:
        mprint("Updating environment script files...\n")
        ectx.save_default_bash_environ()
        ectx.save_default_csh_environ()

    if ctx.invoked_subcommand is None:
        pass
        # click.echo('I was invoked without subcommand')
    else:
        pass
        # click.echo('I am about to invoke %s' % ctx.invoked_subcommand)

def _build_deps_requests(ectx, initial_request):
    assert isinstance(initial_request, InstallationRequest)
    assert isinstance(ectx, EjpmApi)

    install_chain_names = ectx.pm.get_installation_chain_names(initial_request.name, initial_request.deps_only)

    requests = []                       # resulting InstallationRequests
    for name in install_chain_names:

        # Set installation mode to 'single' and 'deps_only' for all except initial request
        if name != initial_request.name:
            mode = 'single'
            deps_only = False
        else:
            mode = initial_request.mode
            deps_only = initial_request.deps_only

        config = {}
        config.update(ectx.db.get_global_config())
        config.update(ectx.db.get_config(name))
        config['build_threads'] = initial_request.config_overrides.get('build_threads', 4)

        # Create installation requrest

        request = InstallationRequest(ectx.pm.recipes_by_name[name],
                                      mode,
                                      config,
                                      initial_request.just_explain,
                                      deps_only)

        # Set right deirectory name in config overwrites
        request.config_overrides['app_path'] = os.path.join(ectx.db.top_dir, name)
        requests.append(request)
    return requests


def _install_packet(ectx, request):
    """Installs packet using its 'installation instruction' class"""

    assert isinstance(request, InstallationRequest)
    assert isinstance(ectx, EjpmApi)

    db = ectx.db
    install_path = os.path.join(db.top_dir, request.name)

    # set_app_path setups parameters (formats all string variables) for this particular path
    request.update_installer_config()
    request.recipe.setup(ectx.db)

    # Pretty header
    mprint("<magenta>=========================================</magenta>")
    mprint("<green> INSTALLING</green> : <blue>{}</blue>", request.name)
    mprint("<magenta>=========================================</magenta>\n")

    # (!) here we actually install the packet
    try:
        request.recipe.step_install()
    except OSError as err:
        mprint("<red>Installation stopped because of the error</red> : {}", err)
        exit(1)

    # if we are here, the packet is installed
    mprint("<green>{} installation step done!</green>\n", request.name)

    # Add to DB that we installed a packet
    mprint("Adding path to database...\n   This {} installation is set as <blue>selected</blue>", request.name)

    from ejpm.engine.db import IS_OWNED, IS_ACTIVE, SOURCE_PATH, BUILD_PATH
    updating_data = {
        IS_OWNED: True,
        IS_ACTIVE: True,
        SOURCE_PATH: request.recipe.source_path,
        BUILD_PATH:  request.recipe.build_path,
        BUILT_WITH_CONFIG: request.recipe.config
    }
    db.update_install(request.recipe.name, request.recipe.install_path, updating_data)
    db.save()


def _install_with_deps(ectx, request):
    assert isinstance(request, InstallationRequest)
    assert isinstance(ectx, EjpmApi)

    must_exist_chain = _build_deps_requests(ectx, request)

    #
    # First. Hit 'setup' function on all dependencies.
    # This will allow us to build the right environment for nonexistent packets
    # If this is just a single packet install it will do no harm
    for request in must_exist_chain:
        # To call setup we need some installation path. Those are dependencies and we only know how to
        # install them to top_dir. So we don't care and set simple os.path.join(...)
        request.config_overrides['app_path'] = os.path.join(ectx.db.top_dir, request.name)
        request.update_installer_config()
        request.recipe.setup(ectx.db)

    #
    # Lets see what is missing and tell it to the user
    missing_chain = []
    mprint("\nCurrent status of the packet and dependencies:")
    for request in must_exist_chain:
        data = ectx.db.get_active_install(request.name)
        if not data:
            mprint("   <blue>{:<9}</blue> : not installed", request.name)
            missing_chain.append(request)
        else:
            is_owned_str = '(owned)' if data['is_owned'] else ''
            mprint("   <blue>{:<9}</blue> : {} {}", request.name, data['install_path'], is_owned_str)

    #
    # Select packets to install. mode tells what we should do with dependencies
    if request.mode == 'missing':
        # select only missing packets
        process_chain = [request for request in must_exist_chain if request in missing_chain]
    elif request.mode == 'single':
        # single = we only one packet
        process_chain = [request]
    elif request.mode == 'all':
        # all - we just overwrite everything
        process_chain = [request for request in must_exist_chain]
    else:
        raise NotImplementedError("installation dependencies mode is not in [missing, single, all]")

    #
    # Is there something to build?
    if not process_chain:
        mprint("Nothing to build and install!")
        return

    # Print user what is going to be built
    mprint("\n <b>INSTALLATION ORDER</b>:")
    for request in process_chain:
        mprint("   <blue>{:<6}</blue> : {}", request.name, request.recipe.install_path)

    # It is just explanation
    if request.just_explain:
        return

    # Set environment before build
    ectx.update_python_env(process_chain, request.mode)  # set environment spitting on existing missing

    #
    for request in process_chain:
        _install_packet(ectx, request)

def _print_help_no_top_path():
    mprint("<red>(!)</red> installation directory is not set <red>(!)</red>\n"
           "ejpm doesn't know where to install missing packets\n\n"
           "<b><blue>what to do:</blue></b>\n"
           "  Provide the top dir to install things to:\n"
           "     ejpm --top-dir=<path to top dir>\n"
           "  Less recommended. Provide each install command with --path flag:\n"
           "     ejpm install <packet> --path=<path for packet>\n"
           "  (--path=... is not just where binary will be installed,\n"
           "   all related stuff is placed there)\n\n"

           "<b><blue>copy&paste:</blue></b>\n"
           "  to install missing packets in this directory: \n"
           "     ejpm --top-dir=`pwd`\n\n"

           "  to install missing packets to your home directory: \n"
           "     ejpm --top-dir=~/.ejana\n\n")